import logging
import os

from .app import name, version
from .camera import AbstractCamera
from .system import AbstractSystem
from flask import Flask, render_template, send_file, Response
from flask_restful import Api, Resource, abort, request
from flask_cors import CORS

from flask.testing import FlaskClient


API_PREFIX = '/api'

class ActionPiAPI(object): 

    def __init__(self, app: Flask, camera: AbstractCamera, system: AbstractSystem):

        self._app = app
        self._api = Api(self._app)

        self._camera = camera
        self._system = system

        #CORS
        CORS(self._app)

        #Setup routes
        self._api.add_resource(Start, API_PREFIX + '/start', resource_class_args=(camera,))
        self._api.add_resource(Stop, API_PREFIX + '/stop', resource_class_args=(camera,))
        self._api.add_resource(Status, API_PREFIX + '/status', resource_class_args=(camera, system))
        self._api.add_resource(Set, API_PREFIX + '/set', resource_class_args=(camera,))
        self._api.add_resource(Hotspot, API_PREFIX + '/hotspot', resource_class_args=(system, ))
        self._api.add_resource(Halt, API_PREFIX + '/halt', resource_class_args=(system,))
        self._api.add_resource(Reboot, API_PREFIX + '/reboot', resource_class_args=(system,))
        self._api.add_resource(Mount, API_PREFIX + '/mount', resource_class_args=(system,))
        self._api.add_resource(Recordings, API_PREFIX + '/recordings', resource_class_args=(camera,))
        self._api.add_resource(Recording, API_PREFIX + '/recording/<string:filename>', resource_class_args=(camera,))

        #Static route
        self._app.add_url_rule('/', '_index', self._index)

        #Preview
        self._app.add_url_rule('/preview', '_preview', self._preview)

    def _index(self):
        return self._app.send_static_file('index.html')    

    def _preview(self):
        frame_buff = self._camera.capture_frame()
        if frame_buff is not None:
            return Response((b'--frame\r\nContent-Type: image/jpeg\r\n\r\n' + frame_buff.read() + b'\r\n'),
                            mimetype='multipart/x-mixed-replace; boundary=frame')
        else:
            return Response('', 409)

    def get_test_client(self) -> FlaskClient:
        """
            Only for test purpose
        """
        return self._app.test_client()

    def get_context(self):
        """
            Only for test purpose
        """
        return self._app.app_context()

    def serve(self):
        logging.info("Serving API (debug: %s)", self._app.config['DEBUG'])
        self._app.run(host=self._app.config['HOST'], port=self._app.config['PORT'], debug=self._app.config['DEBUG'])

    def close(self):
        pass

class Start(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera

    def get(self):
        self._camera.start_recording()
        return '', 204

class Stop(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera
    
    def get(self):
        self._camera.stop_recording()
        return '', 204

#TODO add more logic if needed in future
class Set(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera
    
    def get(self, val: int):
        self._camera.change_framerate(val)

class Status(Resource):
    def __init__(self, camera: AbstractCamera, system: AbstractSystem):
        self._camera = camera
        self._system = system
    
    def get(self):
        return {
            'system': {
                'cpu_temperature': self._system.get_cpu_temp(),
                'cpu_load': self._system.get_cpu_percent(),
                'mem_usage': self._system.get_ram_usage(),
                'disk_usage': self._system.get_disks_usage(),
                'ap_mode': self._system.get_wifi_mode(),
            }, 
            'recording': self._camera.is_recording(),
            'framerate': self._camera.get_framerate()
        }

class Halt(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        logging.info('Shutdown now')
        self._system.halt_system()

class Reboot(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        logging.info('Reboot now')
        self._system.reboot_system()            

class Hotspot(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        enable = request.args.get('enable')
        ssid = request.args.get('ssid')
        password = request.args.get('password')
        country_code = request.args.get('country_code')

        if enable == 'on':
            logging.info('enabling hotspot')
            if not self._system.enable_hotspot(password):
                abort(500)
        elif enable == 'off':
            logging.info('disabling hotspot and enable client mode (AP name: %s, password: %s)', ssid, password)
            if not self._system.connect_to_ap(country_code, ssid, password):
                abort(500)
        else:
            logging.error('enable must be "on" or "off", received: %s', enable)
            return 'enable must be "on" or "off"', 400

        return '', 204

class Mount(Resource):
    def __init__(self, system: AbstractSystem):
        self._system = system

    def get(self):
        mode = request.args.get('mode')
        
        if mode == 'rw':
            logging.info('enabling rw file system')
            self._system.mount_rw()
        elif mode == 'ro':
            logging.info('enabling ro file system')
            self._system.mount_ro()
        else:
            logging.error('mode must be "rw" or "ro", received: %s', mode)
            return 'mode must be "rw" or "ro"', 400

        return '', 204

class Recordings(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera

    def get(self):
        logging.debug('listing files in %s', self._camera.get_output_dir())
        file_names = os.listdir(self._camera.get_output_dir())
        file_names = filter(lambda f: (f is not None) and f.endswith('.h264'), file_names)

        def to_obj(file_name) -> dict:
            return {
                'name': file_name,
                'size': os.path.getsize(os.path.join(self._camera.get_output_dir(), file_name))
            }
        
        return list(map(to_obj, file_names))

class Recording(Resource):
    def __init__(self, camera: AbstractCamera):
        self._camera = camera

    def get(self, filename: str):
        relative_path = os.path.join(self._camera.get_output_dir(), filename)
        absolute_path = os.path.abspath(relative_path)
        logging.debug('downloading file %s', absolute_path)
        if os.path.exists(absolute_path):
            return send_file(absolute_path)
        else:
            return {'message': 'no file in path {} exists'.format(absolute_path)}, 404
        

    


import logging

from flask.app import Config

from .camera import AbstractCamera
from .system import AbstractSystem

class AbstractIO():
    def __init__(self, camera: AbstractCamera, system: AbstractSystem, config: Config):
        self._camera = camera
        self._system = system
        self._config = config
    
    def start_monitoring(self):
        logging.info("Start monitoring GPIO {}".format(self._config['GPIO_SWITCH']))

    def close(self):
        logging.info("Stopping monitoring GPIO {}".format(self._config['GPIO_SWITCH']))


import logging
import time

from threading import Thread, Event, RLock
from flask.app import Config

from .system import AbstractSystem
from .camera import AbstractCamera


class ActionPiWatchdog(object):

    def __init__(self, system: AbstractSystem, camera: AbstractCamera, config: Config):
        self._system = system
        self._camera = camera
        self._config = config
        self._is_watching = Event()
        self._is_triggered = Event()
        self._stop_scheduler = Event()
        self._lock = RLock()
        self._interval = 10
        self._watchdog_not_triggered_interval = self._interval
        self._watchdog_triggered_interval = 120

    def watch(self,interval=10, disks_to_watch=['/']):
        with self._lock:
            self._disks_to_watch = disks_to_watch

            if self._is_watching.is_set():
                logging.warn("Watchdog is already started")
                return

            self._interval = interval
            self._watchdog_not_triggered_interval = self._interval

            class ScheduleThread(Thread):
                @classmethod
                def run(cls):
                    while not self._stop_scheduler.wait(self._interval):
                        self._watchdog_loop()
                    self._stop_scheduler.clear()

            ScheduleThread().start()

            self._is_watching.set()
    
    def _perform_system_status_check(self) -> bool:
        healthy = True

        # Disk usage check
        full_disks = list(filter(lambda disk: (disk['mountpoint'] in self._disks_to_watch) and (disk['percent'] >= self._config['MAX_DISK_USAGE_PERCENT']), self._system.get_disks_usage()))
        if len(full_disks) > 0:
            logging.warning("Disk/s usage of %s is above the allowed maximum %s", full_disks, self._config['MAX_DISK_USAGE_PERCENT'])
            healthy = False
        
        # Temperature check
        if self._system.get_cpu_temp() >= self._config['MAX_CPU_TEMPERATURE_PERCENT']:
            logging.warning("CPU temperature is above the maximum %s/%s", self._system.get_cpu_temp(), self._config['MAX_CPU_TEMPERATURE_PERCENT'])
            healthy = False

        return healthy

    def is_triggered(self) -> bool:
        with self._lock:
            return self._is_triggered.is_set()

    def _watchdog_loop(self):
        if not self._is_triggered.is_set():
            #logging.debug("Watchdog is not triggered perform system status check")
            
            if not self._perform_system_status_check():
                self._is_triggered.set()
                self._camera.stop_recording()
                self._interval = self._watchdog_triggered_interval
        else:
            #logging.debug("Watchdog is triggered perform system status check")
            if self._perform_system_status_check():
                self._is_triggered.clear()
                self._camera.start_recording()
                self._interval = self._watchdog_not_triggered_interval

    
    def is_watching(self):
        with self._lock:
            return self._is_watching.is_set()

    def set_watchdog_triggered_interval(self, interval: int):
        self._watchdog_triggered_interval = interval

    def get_camera(self) -> AbstractCamera:
        return self._camera

    def get_system(self) -> AbstractSystem:
        return self._system

    def unwatch(self):
        with self._lock:
            if self._is_watching.is_set():
                logging.debug("Stopping watchdog...")
                
                self._stop_scheduler.set()
                while self._stop_scheduler.is_set():
                    time.sleep(2)
                
                self._is_watching.clear()
                self._is_triggered.clear()
                self._stop_scheduler.clear()

                logging.info("Watchdog stopped")
            else:
                logging.warn("Watchdog is not started yet")



import logging
import subprocess
import psutil
import os
import re

from io import BytesIO
from pathlib import Path
from time import sleep
from flask.app import Config

from actionpi import AbstractIO, AbstractCamera, AbstractSystem

try:
    from gpiozero import Button
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError("No module gpiozero installed")

try:
    from picamera import PiCamera
except (ImportError, ModuleNotFoundError) as e:
    raise ImportError("No module picamera installed")

WPA_CONFIG_FILE_TEMPLATE=\
"""ctrl_interface=DIR=/var/run/wpa_supplicant GROUP=netdev
update_config=1
country={}

network={{
    ssid="{}"
    psk="{}"
    scan_ssid=1
}}
"""

WPA_MIN_LENGTH_PASSWORD = 8

class RaspberryPiCamera(AbstractCamera):

    def __init__(self,config: Config):
        super().__init__(config)


    def _start(self):
        if self._camera is None:
            self._camera = PiCamera(resolution= (self._config['WIDTH'], self._config['HEIGHT']), framerate=self._config['FRAMERATE'])
            self._camera.rotation = self._config['ROTATION']
            self._camera.start_recording(self._video_file, format='h264')

    def _stop(self):
        if self._camera is not None:
            self._camera.stop_recording()
            self._camera.close()
            self._camera = None

    def _recording(self) -> bool:
        return (self._camera is not None) and (self._camera.recording)

    def _support_split(self) -> bool:
        return True
    
    def _split_recording(self):
        self._camera.split_recording(self._video_file)

    def get_framerate(self) -> int:
        if self._camera is not None:
            return int(self._camera.framerate)
        else:
            return 0
    
    def set_led_status(self, status: bool):
        if self._camera is not None:
            self._camera.led = status

    def capture_frame(self) -> BytesIO:
        if self._camera is not None:
            capture_stream = BytesIO()
            self._camera.capture(capture_stream, 'jpeg', resize=(320, 240), use_video_port=True)
            capture_stream.seek(0)
            return capture_stream

class RaspberryPiSystem(AbstractSystem):
    def get_cpu_temp(self) -> float:
        try:
            temp = int(open('/sys/class/thermal/thermal_zone0/temp').read()) / 1e3
            return temp
        except IOError:
            return 0

    def get_cpu_percent(self) -> int:
        return psutil.cpu_percent(interval=None)

    def get_disks_usage(self) -> list:
        disk_usages = list()
        for part in psutil.disk_partitions(all=False):
            mountpoint = part.mountpoint
            usage = psutil.disk_usage(mountpoint)
            disk_usages.append({
                'mountpoint' : mountpoint,
                'percent' : usage.percent,
                'total' : usage.total,
                'used' : usage.used
            })
        return disk_usages
    
    def get_ram_usage(self) -> int:
        return psutil.virtual_memory().percent

    def halt_system(self):
        subprocess.run(["shutdown", "-H", "now"])

    def reboot_system(self):
        subprocess.run(["shutdown", "-r", "now"])
    
    def get_wifi_mode(self) -> str:
        result = subprocess.run(["/sbin/iwconfig", "wlan0"], universal_newlines=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if result.stdout is not None and len(result.stdout) > 0:
            logging.debug('iwconfig output: {}'.format(result.stdout))
            match = re.search(r'Mode:([-\w]+)\s', result.stdout)
            if match:
                return match.group(1)
            else:
                logging.warn("No match!")
        else:
            logging.error('No output from subprocess (exit code: %d): %s', result.returncode, result.stderr)

        return None
        
    def enable_hotspot(self, password) -> bool:

        if password is not None \
            and isinstance(password, str)  \
            and len(password) >= WPA_MIN_LENGTH_PASSWORD:
            
            logging.info('Using new hotspot password: "{}"'.format(password))
            with open("/boot/wifi_hotspot", "w") as wifi_hotspot_file:
                print(password, file=wifi_hotspot_file, end='')
        else:
            logging.info('Using old password')
            try:
                Path('/boot/wifi_hotspot').unlink()
            except FileNotFoundError:
                logging.warning('Failed to unlink')
            
            Path('/boot/wifi_hotspot').touch()

        try:
            Path('/boot/wifi_client').unlink()
        except FileNotFoundError:
            logging.warning('Failed to unlink')
        
        try:
            Path('/boot/wpa_supplicant.conf').unlink()
        except FileNotFoundError:
            logging.warning('Failed to unlink')
            
        return True

    def connect_to_ap(self, country_code, ssid, password) -> bool:
        try:
            Path('/boot/wifi_hotspot').unlink()
        except FileNotFoundError:
            logging.error('Failed to unlink')

        if(ssid is not None and password is not None and country_code is not None \
            and isinstance(ssid, str) and isinstance(password, str) and isinstance(country_code, str)
            and len(ssid) > 0 and len(password) >= WPA_MIN_LENGTH_PASSWORD) and len(country_code) == 2:
            logging.debug('Using new AP parameter: SSID: "{}"; password: "{}"'.format(ssid, password))
            
            with open("/boot/wpa_supplicant.conf", "w") as wpa_supplicant_file:
                print(WPA_CONFIG_FILE_TEMPLATE.format(country_code, ssid, password), file=wpa_supplicant_file, end='')
        else:
            logging.info('Using old WiFi configuration')

        Path('/boot/wifi_client').touch()

        return True

    def get_hw_revision(self) -> str:
        hw_rev = "00000"
        try:
            f = open('/proc/cpuinfo','r')
            for line in f:
                if line[0:8]=='Revision':
                    hw_rev = line[10:]
            f.close()
        except:
            hw_rev = "00000"

        return hw_rev

    def get_serial(self) -> str:
        # Extract serial from cpuinfo file
        cpuserial = "0000000000000000"
        try:
            f = open('/proc/cpuinfo','r')
            for line in f:
                if line[0:6]=='Serial':
                    cpuserial = line[10:26]
            f.close()
        except:
            cpuserial = "ERROR000000000"

        return cpuserial

    def mount_rw(self):
        Path('/boot/rw').touch()

    def mount_ro(self):
        try:
            Path('/boot/rw').unlink()
        except FileNotFoundError:
            logging.error('Failed to unlink')

    def will_mount_rw(self) -> bool:
        return Path('/boot/rw').exists() and Path('/boot/rw').is_file()

class RaspberryPiIO(AbstractIO):

    def __init__(self, camera: AbstractCamera, system: AbstractSystem, config: Config):
        super().__init__(camera, system, config)
        self.button = Button(config['GPIO_SWITCH'], bounce_time=0.2)
    
    def start_monitoring(self):
        super().start_monitoring()
        self.button.when_pressed = self._start_recording_handler
        self.button.when_released = self._stop_recording_handler
        
        if self.button.is_pressed:
            self._camera.start_recording()

    def _start_recording_handler(self):
        self._camera.start_recording()

    def _stop_recording_handler(self):
        self._camera.stop_recording()
        sleep(1)
        self._system.halt_system()

    def close(self):
        super().close()
        self.button.close()
        self.button = None
